//----------------------------------------------------------------------------------------------
// BSP : Seeed nRF52 Borads 1.1.1
//     : Seeed nRF52 Borads 1.1.8 Confirmed to work 2024/01/06
// Board : Seeed nRF52 Borads / Seeed XIAO nRF52840
//----------------------------------------------------------------------------------------------
// 2024/01/06
//
// nRF52_XIAO   LoRa-E5 mini
//   TX(D6) -----> RX
//   RX(D7) <----- TX
//   3V3    -----> 3V3  Do not supply 5V to LoRa-E5
//   GND    <----> GND

#include <Arduino.h>
#include <Adafruit_GFX.h>
#include <Adafruit_SSD1306.h>
#include "Fonts/tahoma7pt7b.h"

#define HICHG         22        // Charge current　Hi:50mA Lo:100mA       D22 (13u)
#define CHG           23        // Charge LED Lo:ON                       D23 (17u)
#define VBAT_ENABLE   14        // Battery voltage read enable  Lo:enable D14 (14u)   
#define VBAT_READ     32        // Battery voltage read pin               D32 (31u)

#define TIMEOUT    10000        // Response timeout ms
#define LOOP_TIME   5000        // Interval of TX ms

Adafruit_SSD1306 display(128, 64, &Wire, -1);

char MASTER_ADDR[9] = {"52600193"};  // #3 Wio-E5-LE-HF
char SLAVE_ADDR[9] = {"510068D4"};  // #2 Wio-E5 HF
//char RFCFG[64] = {"AT+TEST=RFCFG,922.1,SF7,250,8,8,13,ON,OFF,OFF\r\n"};   // 13dBm=20mW, SF7=11000bps
//char RFCFG[64] = {"AT+TEST=RFCFG,922.1,SF7,250,8,8,0,ON,OFF,OFF\r\n"};   // 0dBm=1mW, SF7=11000bps
char RFCFG[64] = {"AT+TEST=RFCFG,922.6,SF12,125,8,8,13,ON,OFF,OFF\r\n"};   // 13dBm=20mW, SF12=250bps

char recv_buff[512];
bool is_exist = true;
int rssi = 0;
int snr = 0;

void setup(void)
{
  Serial.begin(115200);
//  while (!Serial);
  delay(1000);
  Serial.println("communication test Slave");

  // Pin initialization
//  pinMode(VBAT_READ, INPUT);
  pinMode(CHG, INPUT_PULLUP);       // PULLUP needed to reduce sleep current 
  pinMode(HICHG, OUTPUT);
  pinMode(VBAT_ENABLE, OUTPUT);   
  pinMode(LED_RED, OUTPUT);
  pinMode(LED_GREEN, OUTPUT);
  pinMode(LED_BLUE, OUTPUT);

  digitalWrite(HICHG, LOW);         // Charge current 100mA
  digitalWrite(VBAT_ENABLE, LOW);   // Battery voltage read enable   
  digitalWrite(LED_RED, HIGH);
  digitalWrite(LED_GREEN, HIGH);
  digitalWrite(LED_BLUE, HIGH);

  // ADC initialization　See wireing_analog_nRF52.c:73
  analogReference(AR_DEFAULT);      // Default 0.6V*6=3.6V
  analogReadResolution(12);         // wireing_analog_nRF52.c:39
  analogSampleTime(10);             // default 3uS　3,5,10,15,20,40
  
  // UART initialization
  Serial1.begin(115200);            // Need to match baud rates, LoRa default 9600 
    
  // display initialization
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);
  display.clearDisplay();
  display.setFont(&tahoma7pt7b);
  display.setTextColor(SSD1306_WHITE);
  display.setCursor(0, 15);
  display.println("LoRa-E5");
  display.setCursor(0, 31);
  display.println("P2P Test");
  display.setCursor(0, 47);
  display.println("Slave");  
  display.display();
  delay(2000);
  display.clearDisplay();
    
  // LoRa initialization
  Serial.print("send check response!\r\n");
  at_command("AT\r\n");
  if(!at_response(1, TIMEOUT)) {
    Serial.println("AT timeout");
    is_exist = false;
  }
  at_command("AT+ID\r\n");
  if(!at_response(3, TIMEOUT)) {
    Serial.println("AT+ID timeout");
    is_exist = false;
  }
  at_command("AT+MODE=TEST\r\n");
  if(!at_response(1, TIMEOUT)) {
    Serial.println("AT+MODE=TEST timeout");
    is_exist = false;    
  }
  at_command(RFCFG);
  if(!at_response(1, TIMEOUT)) { 
    Serial.println("AT+TEST=RFCFG timeout");
    is_exist = false;    
  }
  at_command("AT+TEST=?\r\n");
  if(!at_response(2, TIMEOUT)) {
    Serial.println("AT+TEST=? timeout");
    is_exist = false;    
  }
   
  if(is_exist == false) {
    Serial.println("E5 response error");
    display.setCursor(0, 63);
    display.print("E5 no response");
    while(true);
  }
 
  display.setCursor(0, 47);
  display.print("LoRa-E5 Slave");
  display.display();
  delay(2000);
  display.clearDisplay();
}

void loop(void)
{
  uint32_t timestamp = millis();

  // Battery voltage averaging
  int Vadc = 0;
  for (int i = 0; i < 16; i++) 
  {
    Vadc = Vadc + analogRead(VBAT_READ);
  }
  Vadc = Vadc / 16;
  float Vbatt = 2.961 * 3.6 * Vadc / 4096;   // Vref=3.6  ratio=2.961
    
  uint32_t ms = millis();
  Serial.println(ms);

  // data transmit
  char data1[8] = {0,};
  char data2[8] = {0,};
  char cmd[128];
  sprintf(data1, "%08d", ms / 1000);
  sprintf(data2, "%04d", int(Vbatt*1000));
  sprintf(cmd, "AT+TEST=TXLRPKT,\"%s%s%s\"\r\n", SLAVE_ADDR, data1, data2);  // \" is "
  at_command(cmd);
  if(!at_response(2, TIMEOUT)) {
    Serial.println("AT+TEST=TXLRPKT timeout");
  }
    
  // RSSI SNR receive
  at_command("AT+TEST=RXLRPKT\r\n");  
  if(!at_response(3, TIMEOUT)) {
    Serial.println("AT+TEST=RXLRPKT timeout");
  }          
  recv_parse();
  Serial.println(); 
  
  while(millis() - timestamp < LOOP_TIME);     
}


//**************************************************************************************
void at_command(char *p_cmd)
{ 
  // send command
  digitalWrite(LED_RED, LOW);
  delay(10);
  Serial1.print(p_cmd);
  digitalWrite(LED_RED, HIGH);
}

int at_response(int line_numbers, uint32_t timeout)
{
  int timestamp = millis();
  int linefeed = 0;
  char c;
  int index = 0;
  
  // receive response
  digitalWrite(LED_GREEN, LOW);    
    do
    {
      while(Serial1.available() > 0)
      {
        c = Serial1.read();
        recv_buff[index++] = c;
        Serial.print(c);
        if(c == '\n') {
          linefeed++;
          if(linefeed == line_numbers) {    // check response line numbers
            digitalWrite(LED_GREEN, HIGH);
            return 1;
          }
        }
      } 
    } while(linefeed < line_numbers && millis() - timestamp < timeout);  // check timeout       
  digitalWrite(LED_GREEN, HIGH);
  return 0;
}

int recv_parse(void)
{
  char *p_start = NULL;
  char rxdata[32] = {0,};
  char response[32] = {0,};;
  
  sprintf(response, "+TEST: RX \"%s", SLAVE_ADDR);    
  p_start = strstr(recv_buff, response);  
  if (p_start)
  {
    p_start = strstr(recv_buff, SLAVE_ADDR);
    if(!(p_start && (1 == sscanf(p_start + sizeof(SLAVE_ADDR)-1, "%s", rxdata)))) {      
      Serial.println("no addr error");
      return 0;
    }
    p_start = strstr(recv_buff, "RSSI:");
    if(!(p_start && (1 == sscanf(p_start, "RSSI:%d,", &rssi)))) {
      Serial.println("no rssi error");
      return 0;      
    }
    p_start = strstr(recv_buff, "SNR:");
    if(!(p_start && (1 == sscanf(p_start, "SNR:%d", &snr)))) {
      Serial.println("no snr error");
      return 0;
    }

    String s_data = rxdata;
    int sec = (s_data.substring(0, 8)).toInt();  // from x+1 to y
    float Vbatt = (s_data.substring(8, 12)).toInt() / 1000.0;
    
    Serial.print(rssi);
    Serial.print(",");
    Serial.print(snr);
    Serial.print(",");
    Serial.print(Vbatt, 3);
    Serial.print(",");
    Serial.println(sec);
          
    display.clearDisplay();
    display.setCursor(0, 15); display.print("Master");
    display.setCursor(64, 15); display.print(SLAVE_ADDR);
    display.setCursor(0, 31); display.print(rxdata);
    display.setCursor(0, 47); display.print(rssi); display.print(" dBm");
    display.setCursor(64, 47); display.print(snr); display.print(" dB");
    display.setCursor(0, 63); display.print(sec); display.print(" sec");
    display.setCursor(64, 63); display.print(Vbatt, 3); display.print(" V");   
    display.display();   
    return 1;
  } 
  return 0;
}
